@@ -1,11 +1,13 @@
 import json
 import urllib.parse
+import difflib
+import inspect
 
 from tqdm import tqdm
 import geopandas as gpd
 import pandas as pd
 from shapely.ops import transform
-from shapely.geometry import Polygon, LineString, Point, box
+from shapely.geometry import Polygon, MultiPolygon, LineString, Point, box
 from shapely.geometry.base import BaseGeometry
 import matplotlib.pyplot as plt
 import matplotlib.colors as mcolors
@@ -14,7 +16,7 @@ import matplotlib.patches as mpatches
 from mundipy.map import Map
 from mundipy.layer import Layer, VisibleLayer
 from mundipy.api.osm import grab_from_osm
-from mundipy.pcs import choose_pcs
+from mundipy.pcs import choose_pcs, NoProjectionFoundError
 from mundipy.cache import pyproj_transform
 
 class MundiQ:
@@ -67,6 +69,28 @@ class MundiQ:
             # convert to local CRS, because grab_from_osm gives WGS84
             return VisibleLayer(Layer(gdf), self.bbox(), self.center.geometry, pcs=self.pcs)
 
+        # this is an unknown dataset, throw a useful error
+        possible_datasets = list(self.mapdata.collections.keys()) + ['openstreetmap', 'osm']
+        similar_dataset_name = difflib.get_close_matches(dataset, possible_datasets, n=1)
+
+        if len(similar_dataset_name) == 0:
+            raise TypeError('Unknown dataset name %s was passed to Q()' % dataset)
+        else:
+            raise TypeError('Unknown dataset name %s was passed to Q(), did you mean %s?' % (dataset, possible_datasets[0]))
+
+    def call_process(self, fn):
+        # pass dataset as dataframe if requested
+        args = inspect.getfullargspec(fn)[0][1:]
+
+        df_args = []
+        for arg in args:
+            try:
+                df_args.append(self.mapdata.collections[arg])
+            except KeyError:
+                raise TypeError('mundi process() function requests dataset \'%s\', but no dataset was defined on Mundi' % arg)
+
+        return fn(self, *args)
+
     def _bbox(self, distance=500):
         """Builds a bounding box around the center object in the local coordinate system."""
 
@@ -83,7 +107,7 @@ class MundiQ:
 
         # coerce shape into a GeoSeries, no matter what it is
         # shape can literally be anything
-        if isinstance(shape, Polygon):
+        if isinstance(shape, Polygon) or isinstance(shape, MultiPolygon):
             shape = gpd.GeoSeries([shape])
         elif isinstance(shape, gpd.GeoDataFrame):
             shape = shape.geometry
@@ -148,7 +172,7 @@ class Mundi:
         # TODO: drop duplicates, except it's very slow
         #.drop_duplicates(subset=['geometry'])
         Q = MundiQ(self.main.dataframe.iloc[element_index], self.mapdata, plot_target=('geojson' if output_type == 'geojson' else ax), units=self.units, clip_distance=clip_distance)
-        res = fn(Q)
+        res = Q.call_process(fn)
 
         if output_type == 'geojson':
             # convert to dfs
@@ -160,7 +184,7 @@ class Mundi:
             ax.legend(handles=Q.plot_handles, loc='upper right')
             plt.show()
 
-    def q(self, fn, progressbar=False, n_start=0, n_end=0):
+    def q(self, fn, progressbar=False, n_start=None, n_end=None):
         # make iterator unique by geometry
         # TODO: drop duplicates, except it's very slow
         unique_iterator = self.main.dataframe
@@ -170,6 +194,7 @@ class Mundi:
         res_outs = dict()
         # progressbar optional
         finiter = list(unique_iterator.iterrows())[n_start:n_end]
+
         if progressbar:
             finiter = tqdm(finiter, total=len(finiter))
 
@@ -177,8 +202,11 @@ class Mundi:
             # fn(Q) can edit window
             original_shape = window.geometry
 
-            Q = MundiQ(window, self.mapdata)
-            res = fn(Q)
+            try:
+                Q = MundiQ(window, self.mapdata)
+                res = Q.call_process(fn)
+            except NoProjectionFoundError:
+                continue
 
             # if res is None, skip
             if res is None:
@@ -217,4 +245,9 @@ class Mundi:
             if res_shapely_col == 'geometry':
                 res_outs['geometry'].append(original_shape)
 
+        # if res_outs is empty, give a useful error message
+        # creating a GeoDataFrame with an empty array gives an error
+        if len(res_outs) == 0:
+            raise ValueError('all results from mundi.q() process fn were None')
+
         return gpd.GeoDataFrame(res_outs, crs='EPSG:4326', geometry=res_shapely_col)
